//
// PDFsharp - A library for processing PDF
//
// Authors:
//   Klaus Potzesny (mailto:Klaus.Potzesny@pdfsharp.com)
//
// Copyright (c) 2005 empira Software GmbH, Cologne (Germany)
//
// http://www.pdfsharp.com
//
// Permission is hereby granted, free of charge, to any person obtaining
// a copy of this software and associated documentation files (the
// "Software"), to deal in the Software without restriction, including
// without limitation the rights to use, copy, modify, merge, publish,
// distribute, sublicense, and/or sell copies of the Software, and to
// permit persons to whom the Software is furnished to do so, subject to
// the following conditions:
// 
// The above copyright notice and this permission notice shall be
// included in all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
// NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
// LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
// OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
// WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

using System;
using System.Diagnostics;
using System.ComponentModel;
using PdfSharp.Drawing;

namespace PdfSharp.Drawing.BarCodes
{
  /// <summary>
  /// Represents the base class of all bar codes.
  /// </summary>
  public abstract class BarCode
  {
    public BarCode(string text, XSize size, BarCodeDirection direction)
    {
      this.text = text;
      this.size = size;
      this.direction = direction;
    }

    public static BarCode FromType(BarCodeType type, string text, XSize size, BarCodeDirection direction)
    {
      switch (type)
      {
        case BarCodeType.Code2of5Interleaved:
          return new Code2of5Interleaved(text, size, direction);

        case BarCodeType.Code3of9Standard:
          return new Code3of9Standard(text, size, direction);

        default:
          throw new InvalidEnumArgumentException("type", (int)type, typeof(BarCodeType));
      }
    }

    public static BarCode FromType(BarCodeType type, string text, XSize size)
    {
      return FromType(type, text, size, BarCodeDirection.LeftToRight);
    }

    public static BarCode FromType(BarCodeType type, string text)
    {
      return FromType(type, text, XSize.Empty, BarCodeDirection.LeftToRight);
    }

    public static BarCode FromType(BarCodeType type)
    {
      return FromType(type, String.Empty, XSize.Empty, BarCodeDirection.LeftToRight);
    }

    public XSize Size
    {
      get { return this.size; }
      set { this.size = value; }
    }
    protected XSize size;

    /// <summary>
    /// Gets or sets the text the bar code shall represent.
    /// </summary>
    public string Text
    {
      get { return this.text; }
      set
      {
        CheckCode(value);
        this.text = value;
      }
    }
    protected string text;

    /// <summary>
    /// NYI. Always MiddleCenter.
    /// </summary>
    public AnchorType Anchor
    {
      get { return this.anchor; }
      set { this.anchor = value; }
    }
    protected AnchorType anchor;

    public BarCodeDirection Direction
    {
      get { return this.direction; }
      set { this.direction = value; }
    }
    protected BarCodeDirection direction;

    public abstract double WideNarrowRatio { get; set;}

    public TextLocation TextLocation
    {
      get { return this.textLocation; }
      set { this.textLocation = value; }
    }
    protected TextLocation textLocation;

    public int DataLength
    {
      get { return this.dataLength; }
      set { this.dataLength = value; }
    }
    protected int dataLength;

    public char StartChar
    {
      get { return this.startChar; }
      set { this.startChar = value; }
    }
    protected char startChar;

    public char EndChar
    {
      get { return this.endChar; }
      set { this.endChar = value; }
    }
    protected char endChar;

    protected virtual void InitRendering(BarCodeRenderInfo info)
    {
      if (this.text == null)
        throw new InvalidOperationException(BcgSR.BarCodeNotSet);

      if (this.Size.IsEmpty)
        throw new InvalidOperationException(BcgSR.EmptyBarCodeSize);
    }

    protected internal abstract void Render(XGraphics gfx, XBrush brush, XFont font, XPoint center);

    /// <summary>
    /// Whem implemented in a derived class, determines whether the specified string can be used as Text
    /// for this bar code type.
    /// </summary>
    /// <param name="text">The code string to check.</param>
    /// <returns>True if the text can be used for the actual barcode.</returns>
    protected abstract void CheckCode(string text);
  }
}
