//
// PDFsharp - A library for processing PDF
//
// Authors:
//   Stefan Lange (mailto:Stefan.Lange@pdfsharp.com)
//
// Copyright (c) 2005 empira Software GmbH, Cologne (Germany)
//
// http://www.pdfsharp.com
//
// Permission is hereby granted, free of charge, to any person obtaining
// a copy of this software and associated documentation files (the
// "Software"), to deal in the Software without restriction, including
// without limitation the rights to use, copy, modify, merge, publish,
// distribute, sublicense, and/or sell copies of the Software, and to
// permit persons to whom the Software is furnished to do so, subject to
// the following conditions:
// 
// The above copyright notice and this permission notice shall be
// included in all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
// NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
// LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
// OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
// WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

using System;
using System.Diagnostics;
using System.Runtime.InteropServices;
using System.ComponentModel;
using System.IO;
#if Gdip
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Drawing.Imaging;
#endif
#if WinFX
using System.Windows.Media;
#endif
using PdfSharp.Internal;
using PdfSharp.Drawing.Pdf;
using PdfSharp.Fonts.TrueType;
using PdfSharp.Pdf;
using PdfSharp.Pdf.IO;
using PdfSharp.Pdf.Advanced;
using PdfSharp.Pdf.Filters;
using PdfSharp.Pdf.Internal;

namespace PdfSharp.Drawing
{
  /// <summary>
  /// Represents a so called 'PDF form external object', which is typically an imported page of an external
  /// PDF document. XPdfForm objects are used like images to draw an existing PDF page of an external
  /// document in the current document. XPdfForm objects can only be placed in PDF documents. If you try
  /// to draw them using a XGraphics based on an GDI+ context no action is taken if no placeholder image
  /// is specified. Otherwise the place holder is drawn.
  /// </summary>
  public class XPdfForm : XForm
  {
    /// <summary>
    /// Initializes a new instance of the XPdfForm class from the specified path to an external PDF document.
    /// Although PDFsharp internally caches XPdfForm objects it is recommended to reuse PdfXFrom objects
    /// in your code and change the PageNumber property if more than one page is neede form the external
    /// document. Furthermore, because XPdfForm can occupy very much memory, it is recommended to
    /// dispose XPdfForm objects if not needed anymore.
    /// </summary>
    internal XPdfForm(string path)
    {
      int pageNumber;
      path = ExtractPageNumber(path, out pageNumber);

      path = Path.GetFullPath(path);
      if (!File.Exists(path))
        throw new FileNotFoundException(PSSR.FileNotFound(path), path);

      if (PdfReader.TestPdfFile(path) == 0)
        throw new ArgumentException("The specified file has no valid PDF file header.", "path");

      this.path = path;
      if (pageNumber != 0)
        PageNumber = pageNumber;
    }

    /// <summary>
    /// Initializes a new instance of the XPdfForm class such that it can be drawn on the specifed graphics
    /// object. (This constructor will move into the base class in the future.)
    /// </summary>
    /// <param name="gfx">The graphics object that later is used to draw this form.</param>
    /// <param name="size">The size in points of the form.</param>
    public XPdfForm(XGraphics gfx, XSize size)
    {
      if (gfx == null)
        throw new ArgumentNullException("gfx");
      ////// I must tie the XPdfForm to a document immediately, because otherwise I would have no place where
      ////// to store the resources.
      ////if (document == null)
      ////  throw new ArgumentNullException("document", "An XPdfForm template must be associated with a document.");
      if (size.width < 1 || size.height < 1)
        throw new ArgumentNullException("size", "The size of the XPdfForm is to small.");

      this.fromState = FormState.Created;

      // If gfx belongs to a PdfPage also create the pdfFromXObject
      if (gfx.PdfPage != null)
      {
        this.document = gfx.PdfPage.Document;
        this.pdfForm = new PdfFormXObject(document, this);
      }
      this.templateSize = size;
      PdfRectangle rect = new PdfRectangle(XPoint.Empty, size);
      if (this.pdfForm != null)
        this.pdfForm.Elements.SetRectangle(PdfFormXObject.Keys.BBox, rect);
    }

    /// <summary>
    /// Initializes a new instance of the XPdfForm class such that it can be drawn on the specifed graphics
    /// object.
    /// </summary>
    /// <param name="gfx">The graphics object that later is used to draw this form.</param>
    /// <param name="width">The width of the form.</param>
    /// <param name="height">The height of the form.</param>
    public XPdfForm(XGraphics gfx, XUnit width, XUnit height) :
      this(gfx, new XSize(width, height))
    {}

    public XPdfForm(PdfDocument document, XSize size)
    {
      // I must tie the XPdfForm to a document immediately, because otherwise I would have no place where
      // to store the resources.
      if (document == null)
        throw new ArgumentNullException("document", "An XPdfForm template must be associated with a document.");
      if (size.width < 1 || size.height < 1)
        throw new ArgumentNullException("size", "The size of the XPdfForm is to small.");

      this.fromState = FormState.Created;
      this.document = document;
      this.pdfForm = new PdfFormXObject(document, this);
      this.templateSize = size;
      PdfRectangle rect = new PdfRectangle(XPoint.Empty, size);
      this.pdfForm.Elements.SetRectangle(PdfFormXObject.Keys.BBox, rect);
    }

    public XPdfForm(PdfDocument document, XUnit width, XUnit height) :
      this(document, new XSize(width, height))
    {}

    public static new XPdfForm FromFile(string path)
    {
      // TODO: Same file should return same object (thats why the function is static).
      return new XPdfForm(path);
    }

    void Initialize()
    {
      // ImageFormat has no overridden Equals...
    }

    /// <summary>
    /// This function should be called when drawing the content of this form is finished.
    /// The XGraphics object used for drawing the content is disposed by this function and 
    /// cannot be used for any further drawing operations.
    /// PDFsharp automatically calls this function when this form was used the first time
    /// in a DrawImage function. 
    /// </summary>
    public void DrawingFinished()
    {
      if (this.fromState == FormState.Finished)
        return;

      if (this.fromState == FormState.NotATemplate)
        throw new InvalidOperationException("This object is an imported PDF page and you cannot finish drawing on it because you must not draw on it at all.");

      Finish();
    }

    /// <summary>
    /// Called from XGraphics constructor that creates an instance that work on this form.
    /// </summary>
    internal void AssociateGraphics(XGraphics gfx)
    {
      if (this.fromState == FormState.NotATemplate)
        throw new NotImplementedException("The current version of PDFsharp cannot draw on an imported page.");

      if (this.fromState == FormState.UnderConstruction)
        throw new InvalidOperationException("An XGraphics object already exists for this form.");

      if (this.fromState == FormState.Finished)
        throw new InvalidOperationException("After drawing a form it cannot be modified anymore.");

      Debug.Assert(this.fromState == FormState.Created);
      this.fromState = FormState.UnderConstruction;
      this.gfx = gfx;
    }
    internal XGraphics gfx;

    /// <summary>
    /// Sets the form in the state FormState.Finished.
    /// </summary>
    internal void Finish()
    {
      if (this.fromState == FormState.NotATemplate || this.fromState == FormState.Finished)
        return;

      if (this.gfx.metafile != null)
        this.image = this.gfx.metafile;

      Debug.Assert(this.fromState == FormState.Created || this.fromState == FormState.UnderConstruction);
      this.fromState = FormState.Finished;
      this.gfx.Dispose();
      this.gfx = null;

      if (this.pdfRenderer != null)
      {
        this.pdfForm.Stream = new PdfDictionary.PdfStream(PdfEncoders.RawEncoding.GetBytes(this.pdfRenderer.GetContent()), this.pdfForm);

        if (this.document.Options.CompressContentStreams)
        {
          this.pdfForm.Stream.Value = Filtering.FlateDecode.Encode(this.pdfForm.Stream.Value);
          this.pdfForm.Elements["/Filter"] = new PdfName("/FlateDecode");
        }
        int length = this.pdfForm.Stream.Length;
        this.pdfForm.Elements.SetInteger("/Length", length);
      }
    }

    internal bool IsTemplate
    {
      get { return this.fromState != FormState.NotATemplate; }
    }
    FormState fromState;

    internal PdfDocument Document
    {
      get { return this.document; }
    }
    PdfDocument document;

    /// <summary>
    /// TODO:
    /// Frees the memory occupied by the underlying imported PDF document, even if other XPdfForm objects
    /// refer to this document. A reuse of this object doesnt fail, because the underlying PDF document
    /// is re-imported if neccessary. NYI
    /// </summary>
    protected override void Dispose(bool disposing)
    {
      // TODO: remove from thread local cache and from all other XPdfForm objects.
      if (!this.disposed)
      {
        try
        {
          if (disposing)
          {
            //...
          }
          //...
        }
        finally
        {
          base.Dispose(disposing);
        }
      }
      this.disposed = true;
    }
    bool disposed;

    /// <summary>
    /// Gets or sets an image that is used for drawing if the current XGraphics object cannot handle
    /// PDF forms. A place holder is useful for showing a preview of a page on the display, because
    /// PDFsharp cannot render native PDF objects.
    /// </summary>
    public XImage PlaceHolder
    {
      get {return this.placeHolder;}
      set {this.placeHolder = value;}
    }
    XImage placeHolder;

    /// <summary>
    /// Gets the number of pages in the PDF form.
    /// </summary>
    public int PageCount
    {
      get 
      { 
        if (IsTemplate)
          return 1;
        if (this.pageCount == -1)
          this.pageCount = ExternalDocument.Pages.Count;
        return this.pageCount; 
      }
    }
    int pageCount = -1;

    /// <summary>
    /// Get the width of the page identified by the property PageNumber.
    /// </summary>
    public override int Width
    {
      get 
      {
        if (IsTemplate)
          return (int)this.templateSize.width;

        PdfPage page = ExternalDocument.Pages[this.pageNumber - 1];
        return (int)page.Width;
      }
    }

    /// <summary>
    /// Get the width of the page identified by the property PageNumber.
    /// </summary>
    public override int Height
    {
      get 
      {
        if (IsTemplate)
          return (int)this.templateSize.height;

        PdfPage page = ExternalDocument.Pages[this.pageNumber - 1];
        return (int)page.Height;
      }
    }

    public override XSize Size
    {
      get
      {
        if (IsTemplate)
          return this.templateSize;

        PdfPage page = ExternalDocument.Pages[this.pageNumber - 1];
        return new XSize(page.Width, page.Height);
      }
    }
    XSize templateSize;

    /// <summary>
    /// Gets 72, the horizontal resolution by design of a form X object.
    /// </summary>
    public override double HorizontalResolution
    {
      get {return 72;}
    }

    /// <summary>
    /// Gets 72, the vertical resolution by design of a form X object.
    /// </summary>
    public override double VerticalResolution
    {
      get {return 72;}
    }

    public XRect BoundingBox
    {
      get { return this.boundingBox; }
      set { this.boundingBox = value; }  // TODO: pdfForm = null
    }
    XRect boundingBox;

    public XMatrix Transform
    {
      get { return this.transform; }
      set 
      {
        if (IsTemplate)
        {
          if (this.fromState == FormState.Finished)
            throw new InvalidOperationException("After a XPdfForm was once drawn it must not be modified.");
          this.transform = value;
          return;
        }

        if (this.transform != value)
        {
          // discard PdfFromXObject when Transform changed
          this.pdfForm = null;
          this.transform = value; 
        }
      }
    }
    XMatrix transform = XMatrix.Identity;

    /// <summary>
    /// Gets or sets the page number in the external PDF document this object refers to. The page number
    /// is one-based, i.e. it is in the range from 1 to PageCount. The default value is 1.
    /// </summary>
    public int PageNumber
    {
      get { return this.pageNumber; }
      set 
      { 
        if (IsTemplate)
          throw new InvalidOperationException("The page number of an XPdfForm template cannot be modified.");

        if (this.pageNumber != value)
        {
          this.pageNumber = value;
          // dispose PdfFromXObject when number has changed
          this.pdfForm = null;
        }
      }
    }
    int pageNumber = 1;

    /// <summary>
    /// Gets the underlying document from which pages are imported.
    /// </summary>
    internal PdfDocument ExternalDocument
    {
      // The problem is that you can ask an XPdfForm about the number of its pages before it was
      // drawn the first time. At this moment the XPdfForm doesnt know the document where it will
      // be later draw on one of its pages. To prevent the import of the same document more than
      // once, all imported documents of a thread are cached. The cache is local to the current 
      // thread and not to the appdomain, because I wont get problems in a multi-thread environment
      // that I dont understand.
      get
      {
        if (IsTemplate)
          throw new InvalidOperationException("This XPdfForm is a template and not an imported PDF page; therefore it has no external document..");

        if (this.externalDocument == null)
          this.externalDocument = PdfDocument.Tls.GetDocument(path);
        return this.externalDocument;
      }
    }
    internal PdfDocument externalDocument;

    /// <summary>
    /// Extracts the page number if the path has the form 'MyFile.pdf#123' and returns
    /// the actual path without the number sign and the following digits.
    /// </summary>
    public static string ExtractPageNumber(string path, out int pageNumber)
    {
      if (path == null)
        throw new ArgumentNullException("path");

      pageNumber = 0;
      int length = path.Length;
      if (length != 0)
      {
        length--;
        if (Char.IsDigit(path, length))
        {
          while(Char.IsDigit(path, length) && length >= 0)
            length--;
          if (length > 0 && path[length] == '#')
          {
            // must have at least one dot left of colon to distinguish from e.g. '#123'
            if (path.IndexOf('.') != -1)
            {
              pageNumber = Int32.Parse(path.Substring(length + 1));
              path = path.Substring(0, length);
            }
          }
        }
      }
      return path;
    }

    internal PdfResources Resources
    {
      get
      {
        Debug.Assert(IsTemplate, "This function is for form templates only.");
        if (this.resources == null)
          this.resources = (PdfResources)this.pdfForm.Elements.GetValue(PdfFormXObject.Keys.Resources, VCF.CreateIndirect);
        return this.resources;
      }
    }
    PdfResources resources;

    /// <summary>
    /// Gets the resource name of the specified font within this form.
    /// </summary>
    internal string GetFontName(XFont font)
    {
      PdfFont pdfFont = this.document.FontTable.GetFont(font);
      Debug.Assert(pdfFont != null);
      string name = Resources.AddFont(pdfFont);
      return name;
    }

    /// <summary>
    /// Gets the resource name of the specified image within this form.
    /// </summary>
    internal string GetImageName(XImage image)
    {
      PdfImage pdfImage = this.document.ImageTable.GetImage(image);
      Debug.Assert(pdfImage != null);
      string name = Resources.AddImage(pdfImage);
      return name;
    }

    internal PdfFormXObject PdfForm
    {
      get
      {
        if (this.pdfForm.Reference == null)
          this.document.irefTable.Add(this.pdfForm);
        return this.pdfForm;
      }
    }

    /// <summary>
    /// Gets the resource name of the specified form within this form.
    /// </summary>
    internal string GetFormName(XPdfForm form)
    {
      PdfFormXObject pdfForm = this.document.FormTable.GetForm(form);
      Debug.Assert(pdfForm != null);
      string name = Resources.AddForm(pdfForm);
      return name;
    }

    // HACK: ImportedObjectTable here prevents to get reused from other documents
    // TODO: Rewrite PdfFormXObject ctor to get ImportedObjectTable from current document
    internal ImportedObjectTable importDocument;

    /// <summary>
    /// The PdfFormXObject gets invalid when PageNumber or transform changed. This is because a modification
    /// of an XPdfForm must not change objects that are already been drawn.
    /// </summary>
    internal PdfFormXObject pdfForm;

    internal XGraphicsPdfRenderer pdfRenderer;
  }
}
